@@ -46,12 +46,11 @@
 #include "crypto/s2k.h"
 #include "fingerprint.h"
 #include "pgp-key.h"
-#include "list.h"
 #include "crypto.h"
-#include "json_utils.h"
+#include "json-utils.h"
 #include <algorithm>
 
-static pgp_map_t packet_tag_map[] = {
+static const id_str_pair packet_tag_map[] = {
   {PGP_PKT_RESERVED, "Reserved"},
   {PGP_PKT_PK_SESSION_KEY, "Public-Key Encrypted Session Key"},
   {PGP_PKT_SIGNATURE, "Signature"},
@@ -73,11 +72,10 @@ static pgp_map_t packet_tag_map[] = {
   {PGP_PKT_SE_IP_DATA, "Symmetric Encrypted and Integrity Protected Data"},
   {PGP_PKT_MDC, "Modification Detection Code"},
   {PGP_PKT_AEAD_ENCRYPTED, "AEAD Encrypted Data Packet"},
-
-  {0x00, NULL}, /* this is the end-of-array marker */
+  {0x00, NULL},
 };
 
-static pgp_map_t sig_type_map[] = {
+static const id_str_pair sig_type_map[] = {
   {PGP_SIG_BINARY, "Signature of a binary document"},
   {PGP_SIG_TEXT, "Signature of a canonical text document"},
   {PGP_SIG_STANDALONE, "Standalone signature"},
@@ -93,10 +91,10 @@ static pgp_map_t sig_type_map[] = {
   {PGP_SIG_REV_CERT, "Certification revocation signature"},
   {PGP_SIG_TIMESTAMP, "Timestamp signature"},
   {PGP_SIG_3RD_PARTY, "Third-Party Confirmation signature"},
-  {0x00, NULL}, /* this is the end-of-array marker */
+  {0x00, NULL},
 };
 
-static pgp_map_t sig_subpkt_type_map[] = {
+static const id_str_pair sig_subpkt_type_map[] = {
   {PGP_SIG_SUBPKT_CREATION_TIME, "signature creation time"},
   {PGP_SIG_SUBPKT_EXPIRATION_TIME, "signature expiration time"},
   {PGP_SIG_SUBPKT_EXPORT_CERT, "exportable certification"},
@@ -122,10 +120,10 @@ static pgp_map_t sig_subpkt_type_map[] = {
   {PGP_SIG_SUBPKT_EMBEDDED_SIGNATURE, "embedded signature"},
   {PGP_SIG_SUBPKT_ISSUER_FPR, "issuer fingerprint"},
   {PGP_SIG_SUBPKT_PREFERRED_AEAD, "preferred AEAD algorithms"},
-  {0x00, NULL}, /* this is the end-of-array marker */
+  {0x00, NULL},
 };
 
-static pgp_map_t key_type_map[] = {
+static const id_str_pair key_type_map[] = {
   {PGP_PKT_SECRET_KEY, "Secret key"},
   {PGP_PKT_PUBLIC_KEY, "Public key"},
   {PGP_PKT_SECRET_SUBKEY, "Secret subkey"},
@@ -133,7 +131,7 @@ static pgp_map_t key_type_map[] = {
   {0x00, NULL},
 };
 
-static pgp_map_t pubkey_alg_map[] = {
+static const id_str_pair pubkey_alg_map[] = {
   {PGP_PKA_RSA, "RSA (Encrypt or Sign)"},
   {PGP_PKA_RSA_ENCRYPT_ONLY, "RSA (Encrypt-Only)"},
   {PGP_PKA_RSA_SIGN_ONLY, "RSA (Sign-Only)"},
@@ -145,10 +143,10 @@ static pgp_map_t pubkey_alg_map[] = {
   {PGP_PKA_RESERVED_DH, "Reserved for DH (X9.42)"},
   {PGP_PKA_EDDSA, "EdDSA"},
   {PGP_PKA_SM2, "SM2"},
-  {0x00, NULL}, /* this is the end-of-array marker */
+  {0x00, NULL},
 };
 
-static pgp_map_t symm_alg_map[] = {
+static const id_str_pair symm_alg_map[] = {
   {PGP_SA_PLAINTEXT, "Plaintext"},
   {PGP_SA_IDEA, "IDEA"},
   {PGP_SA_TRIPLEDES, "TripleDES"},
@@ -162,10 +160,10 @@ static pgp_map_t symm_alg_map[] = {
   {PGP_SA_CAMELLIA_192, "Camellia-192"},
   {PGP_SA_CAMELLIA_256, "Camellia-256"},
   {PGP_SA_SM4, "SM4"},
-  {0x00, NULL}, /* this is the end-of-array marker */
+  {0x00, NULL},
 };
 
-static pgp_map_t hash_alg_map[] = {
+static const id_str_pair hash_alg_map[] = {
   {PGP_HASH_MD5, "MD5"},
   {PGP_HASH_SHA1, "SHA1"},
   {PGP_HASH_RIPEMD, "RIPEMD160"},
@@ -176,25 +174,25 @@ static pgp_map_t hash_alg_map[] = {
   {PGP_HASH_SM3, "SM3"},
   {PGP_HASH_SHA3_256, "SHA3-256"},
   {PGP_HASH_SHA3_512, "SHA3-512"},
-  {0x00, NULL}, /* this is the end-of-array marker */
+  {0x00, NULL},
 };
 
-static pgp_map_t z_alg_map[] = {
+static const id_str_pair z_alg_map[] = {
   {PGP_C_NONE, "Uncompressed"},
   {PGP_C_ZIP, "ZIP"},
   {PGP_C_ZLIB, "ZLib"},
   {PGP_C_BZIP2, "BZip2"},
-  {0x00, NULL}, /* this is the end-of-array marker */
+  {0x00, NULL},
 };
 
-static pgp_map_t aead_alg_map[] = {
+static const id_str_pair aead_alg_map[] = {
   {PGP_AEAD_NONE, "None"},
   {PGP_AEAD_EAX, "EAX"},
   {PGP_AEAD_OCB, "OCB"},
-  {0x00, NULL}, /* this is the end-of-array marker */
+  {0x00, NULL},
 };
 
-static pgp_map_t revoc_reason_map[] = {
+static const id_str_pair revoc_reason_map[] = {
   {PGP_REVOCATION_NO_REASON, "No reason"},
   {PGP_REVOCATION_SUPERSEDED, "Superseded"},
   {PGP_REVOCATION_COMPROMISED, "Compromised"},
@@ -325,7 +323,7 @@ dst_print_mpi(pgp_dest_t *dst, const char *name, pgp_mpi_t *mpi, bool dumpbin)
 static void
 dst_print_palg(pgp_dest_t *dst, const char *name, pgp_pubkey_alg_t palg)
 {
-    const char *palg_name = pgp_str_from_map(palg, pubkey_alg_map);
+    const char *palg_name = id_str_pair::lookup(pubkey_alg_map, palg, "Unknown");
     if (!name) {
         name = "public key algorithm";
     }
@@ -336,7 +334,7 @@ dst_print_palg(pgp_dest_t *dst, const char *name, pgp_pubkey_alg_t palg)
 static void
 dst_print_halg(pgp_dest_t *dst, const char *name, pgp_hash_alg_t halg)
 {
-    const char *halg_name = pgp_str_from_map(halg, hash_alg_map);
+    const char *halg_name = id_str_pair::lookup(hash_alg_map, halg, "Unknown");
     if (!name) {
         name = "hash algorithm";
     }
@@ -347,7 +345,7 @@ dst_print_halg(pgp_dest_t *dst, const char *name, pgp_hash_alg_t halg)
 static void
 dst_print_salg(pgp_dest_t *dst, const char *name, pgp_symm_alg_t salg)
 {
-    const char *salg_name = pgp_str_from_map(salg, symm_alg_map);
+    const char *salg_name = id_str_pair::lookup(symm_alg_map, salg, "Unknown");
     if (!name) {
         name = "symmetric algorithm";
     }
@@ -358,7 +356,7 @@ dst_print_salg(pgp_dest_t *dst, const char *name, pgp_symm_alg_t salg)
 static void
 dst_print_aalg(pgp_dest_t *dst, const char *name, pgp_aead_alg_t aalg)
 {
-    const char *aalg_name = pgp_str_from_map(aalg, aead_alg_map);
+    const char *aalg_name = id_str_pair::lookup(aead_alg_map, aalg, "Unknown");
     if (!name) {
         name = "aead algorithm";
     }
@@ -369,7 +367,7 @@ dst_print_aalg(pgp_dest_t *dst, const char *name, pgp_aead_alg_t aalg)
 static void
 dst_print_zalg(pgp_dest_t *dst, const char *name, pgp_compression_type_t zalg)
 {
-    const char *zalg_name = pgp_str_from_map(zalg, z_alg_map);
+    const char *zalg_name = id_str_pair::lookup(z_alg_map, zalg, "Unknown");
     if (!name) {
         name = "compression algorithm";
     }
@@ -386,7 +384,8 @@ dst_print_raw(pgp_dest_t *dst, const char *name, const void *data, size_t len)
 }
 
 static void
-dst_print_algs(pgp_dest_t *dst, const char *name, uint8_t *algs, size_t algc, pgp_map_t map[])
+dst_print_algs(
+  pgp_dest_t *dst, const char *name, uint8_t *algs, size_t algc, const id_str_pair map[])
 {
     if (!name) {
         name = "algorithms";
@@ -394,7 +393,8 @@ dst_print_algs(pgp_dest_t *dst, const char *name, uint8_t *algs, size_t algc, pg
 
     dst_printf(dst, "%s: ", name);
     for (size_t i = 0; i < algc; i++) {
-        dst_printf(dst, "%s%s", pgp_str_from_map(algs[i], map), i + 1 < algc ? ", " : "");
+        dst_printf(
+          dst, "%s%s", id_str_pair::lookup(map, algs[i], "Unknown"), i + 1 < algc ? ", " : "");
     }
     dst_printf(dst, " (");
     for (size_t i = 0; i < algc; i++) {
@@ -406,7 +406,7 @@ dst_print_algs(pgp_dest_t *dst, const char *name, uint8_t *algs, size_t algc, pg
 static void
 dst_print_sig_type(pgp_dest_t *dst, const char *name, pgp_sig_type_t sigtype)
 {
-    const char *sig_name = pgp_str_from_map(sigtype, sig_type_map);
+    const char *sig_name = id_str_pair::lookup(sig_type_map, sigtype, "Unknown");
     if (!name) {
         name = "signature type";
     }
@@ -532,7 +532,7 @@ static void         stream_dump_signature_pkt(rnp_dump_ctx_t * ctx,
 static void
 signature_dump_subpacket(rnp_dump_ctx_t *ctx, pgp_dest_t *dst, const pgp_sig_subpkt_t &subpkt)
 {
-    const char *sname = pgp_str_from_map(subpkt.type, sig_subpkt_type_map);
+    const char *sname = id_str_pair::lookup(sig_subpkt_type_map, subpkt.type, "Unknown");
 
     switch (subpkt.type) {
     case PGP_SIG_SUBPKT_CREATION_TIME:
@@ -577,8 +577,22 @@ signature_dump_subpacket(rnp_dump_ctx_t *ctx, pgp_dest_t *dst, const pgp_sig_sub
     case PGP_SIG_SUBPKT_ISSUER_KEY_ID:
         dst_print_hex(dst, sname, subpkt.fields.issuer, PGP_KEY_ID_SIZE, false);
         break;
-    case PGP_SIG_SUBPKT_NOTATION_DATA:
+    case PGP_SIG_SUBPKT_NOTATION_DATA: {
+        std::string          name(subpkt.fields.notation.name,
+                         subpkt.fields.notation.name + subpkt.fields.notation.nlen);
+        std::vector<uint8_t> value(subpkt.fields.notation.value,
+                                   subpkt.fields.notation.value + subpkt.fields.notation.vlen);
+        if (subpkt.fields.notation.human) {
+            dst_printf(dst, "%s: %s = ", sname, name.c_str());
+            dst_printf(dst, "%.*s\n", (int) value.size(), (char *) value.data());
+        } else {
+            char hex[64];
+            vsnprinthex(hex, sizeof(hex), value.data(), value.size());
+            dst_printf(dst, "%s: %s = ", sname, name.c_str());
+            dst_printf(dst, "0x%s (%zu bytes)\n", hex, value.size());
+        }
         break;
+    }
     case PGP_SIG_SUBPKT_PREFERRED_HASH:
         dst_print_algs(dst,
                        "preferred hash algorithms",
@@ -626,7 +640,7 @@ signature_dump_subpacket(rnp_dump_ctx_t *ctx, pgp_dest_t *dst, const pgp_sig_sub
         break;
     case PGP_SIG_SUBPKT_REVOCATION_REASON: {
         int         code = subpkt.fields.revocation_reason.code;
-        const char *reason = pgp_str_from_map(code, revoc_reason_map);
+        const char *reason = id_str_pair::lookup(revoc_reason_map, code, "Unknown");
         dst_printf(dst, "%s: %d (%s)\n", sname, code, reason);
         dst_print_raw(dst,
                       "message",
@@ -799,7 +813,7 @@ stream_dump_key(rnp_dump_ctx_t *ctx, pgp_source_t *src, pgp_dest_t *dst)
         return ret;
     }
 
-    dst_printf(dst, "%s packet\n", pgp_str_from_map(key.tag, key_type_map));
+    dst_printf(dst, "%s packet\n", id_str_pair::lookup(key_type_map, key.tag, "Unknown"));
     indent_dest_increase(dst);
 
     dst_printf(dst, "version: %d\n", (int) key.version);
@@ -1225,7 +1239,7 @@ stream_dump_packets_raw(rnp_dump_ctx_t *ctx, pgp_source_t *src, pgp_dest_t *dst)
     /* do not allow endless recursion */
     if (++ctx->layers > MAXIMUM_NESTING_LEVEL) {
         RNP_LOG("Too many OpenPGP nested layers during the dump.");
-        dst_printf(dst, ":too many OpenPGP packet layers, stopping.");
+        dst_printf(dst, ":too many OpenPGP packet layers, stopping.\n");
         ret = RNP_SUCCESS;
         goto finish;
     }
@@ -1300,15 +1314,18 @@ stream_dump_packets_raw(rnp_dump_ctx_t *ctx, pgp_source_t *src, pgp_dest_t *dst)
         case PGP_PKT_SE_DATA:
         case PGP_PKT_SE_IP_DATA:
         case PGP_PKT_AEAD_ENCRYPTED:
+            ctx->stream_pkts++;
             ret = stream_dump_encrypted(src, dst, hdr.tag);
             break;
         case PGP_PKT_ONE_PASS_SIG:
             ret = stream_dump_one_pass(src, dst);
             break;
         case PGP_PKT_COMPRESSED:
+            ctx->stream_pkts++;
             ret = stream_dump_compressed(ctx, src, dst);
             break;
         case PGP_PKT_LITDATA:
+            ctx->stream_pkts++;
             ret = stream_dump_literal(src, dst);
             break;
         case PGP_PKT_MARKER:
@@ -1328,6 +1345,13 @@ stream_dump_packets_raw(rnp_dump_ctx_t *ctx, pgp_source_t *src, pgp_dest_t *dst)
             RNP_LOG("failed to process packet");
             goto finish;
         }
+
+        if (ctx->stream_pkts > MAXIMUM_STREAM_PKTS) {
+            RNP_LOG("Too many OpenPGP stream packets during the dump.");
+            dst_printf(dst, ":too many OpenPGP stream packets, stopping.\n");
+            ret = RNP_SUCCESS;
+            goto finish;
+        }
     }
 
     ret = RNP_SUCCESS;
@@ -1369,6 +1393,7 @@ stream_dump_packets(rnp_dump_ctx_t *ctx, pgp_source_t *src, pgp_dest_t *dst)
     rnp_result_t ret = RNP_ERROR_GENERIC;
 
     ctx->layers = 0;
+    ctx->stream_pkts = 0;
     /* check whether source is cleartext - then skip till the signature */
     if (is_cleartext_source(src)) {
         dst_printf(dst, ":cleartext signed data\n");
@@ -1400,11 +1425,9 @@ stream_dump_packets(rnp_dump_ctx_t *ctx, pgp_source_t *src, pgp_dest_t *dst)
         goto finish;
     }
     indent = true;
-
     indent_dest_set(&wrdst, 0);
 
     ret = stream_dump_packets_raw(ctx, src, &wrdst);
-
 finish:
     if (armored) {
         src_close(&armorsrc);
@@ -1416,7 +1439,7 @@ finish:
 }
 
 static bool
-obj_add_intstr_json(json_object *obj, const char *name, int val, pgp_map_t map[])
+obj_add_intstr_json(json_object *obj, const char *name, int val, const id_str_pair map[])
 {
     if (!obj_add_field_json(obj, name, json_object_new_int(val))) {
         return false;
@@ -1425,7 +1448,7 @@ obj_add_intstr_json(json_object *obj, const char *name, int val, pgp_map_t map[]
         return true;
     }
     char        namestr[64] = {0};
-    const char *str = pgp_str_from_map(val, map);
+    const char *str = id_str_pair::lookup(map, val, "Unknown");
     snprintf(namestr, sizeof(namestr), "%s.str", name);
     return obj_add_field_json(obj, namestr, json_object_new_string(str));
 }
@@ -1447,7 +1470,7 @@ obj_add_mpi_json(json_object *obj, const char *name, const pgp_mpi_t *mpi, bool
 
 static bool
 subpacket_obj_add_algs(
-  json_object *obj, const char *name, uint8_t *algs, size_t len, pgp_map_t map[])
+  json_object *obj, const char *name, uint8_t *algs, size_t len, const id_str_pair map[])
 {
     json_object *jso_algs = json_object_new_array();
     if (!jso_algs || !obj_add_field_json(obj, name, jso_algs)) {
@@ -1470,8 +1493,9 @@ subpacket_obj_add_algs(
         return false;
     }
     for (size_t i = 0; i < len; i++) {
-        if (!array_add_element_json(jso_algs,
-                                    json_object_new_string(pgp_str_from_map(algs[i], map)))) {
+        if (!array_add_element_json(
+              jso_algs,
+              json_object_new_string(id_str_pair::lookup(map, algs[i], "Unknown")))) {
             return false;
         }
     }
@@ -1683,7 +1707,23 @@ signature_dump_subpacket_json(rnp_dump_ctx_t *        ctx,
     case PGP_SIG_SUBPKT_ISSUER_FPR:
         return obj_add_hex_json(
           obj, "fingerprint", subpkt.fields.issuer_fp.fp, subpkt.fields.issuer_fp.len);
-    case PGP_SIG_SUBPKT_NOTATION_DATA:
+    case PGP_SIG_SUBPKT_NOTATION_DATA: {
+        bool human = subpkt.fields.notation.human;
+        if (!json_add(obj, "human", human) || !json_add(obj,
+                                                        "name",
+                                                        (char *) subpkt.fields.notation.name,
+                                                        subpkt.fields.notation.nlen)) {
+            return false;
+        }
+        if (human) {
+            return json_add(obj,
+                            "value",
+                            (char *) subpkt.fields.notation.value,
+                            subpkt.fields.notation.vlen);
+        }
+        return obj_add_hex_json(
+          obj, "value", subpkt.fields.notation.value, subpkt.fields.notation.vlen);
+    }
     default:
         if (!ctx->dump_packets) {
             return obj_add_hex_json(obj, "raw", subpkt.data, subpkt.len);
@@ -2386,15 +2426,18 @@ stream_dump_raw_packets_json(rnp_dump_ctx_t *ctx, pgp_source_t *src, json_object
         case PGP_PKT_SE_DATA:
         case PGP_PKT_SE_IP_DATA:
         case PGP_PKT_AEAD_ENCRYPTED:
+            ctx->stream_pkts++;
             ret = stream_dump_encrypted_json(src, pkt, hdr.tag);
             break;
         case PGP_PKT_ONE_PASS_SIG:
             ret = stream_dump_one_pass_json(src, pkt);
             break;
         case PGP_PKT_COMPRESSED:
+            ctx->stream_pkts++;
             ret = stream_dump_compressed_json(ctx, src, pkt);
             break;
         case PGP_PKT_LITDATA:
+            ctx->stream_pkts++;
             ret = stream_dump_literal_json(src, pkt);
             break;
         case PGP_PKT_MARKER:
@@ -2417,6 +2460,12 @@ stream_dump_raw_packets_json(rnp_dump_ctx_t *ctx, pgp_source_t *src, json_object
             ret = RNP_ERROR_OUT_OF_MEMORY;
             goto done;
         }
+        if (ctx->stream_pkts > MAXIMUM_STREAM_PKTS) {
+            RNP_LOG("Too many OpenPGP stream packets during the dump.");
+            ret = RNP_SUCCESS;
+            goto done;
+        }
+
         pkt = NULL;
     }
 done:
