@@ -48,7 +48,6 @@
 #include "stream-packet.h"
 #include "stream-armor.h"
 #include "stream-sig.h"
-#include "list.h"
 #include "pgp-key.h"
 #include "fingerprint.h"
 #include "types.h"
@@ -90,7 +89,7 @@ typedef struct pgp_dest_encrypted_param_t {
     bool                    has_mdc; /* encrypted with mdc, i.e. tag 18 */
     bool                    aead;    /* we use AEAD encryption */
     pgp_crypt_t             encrypt; /* encrypting crypto */
-    pgp_hash_t              mdc;     /* mdc SHA1 hash */
+    rnp::Hash               mdc;     /* mdc SHA1 hash */
     pgp_aead_alg_t          aalg;    /* AEAD algorithm used */
     uint8_t                 iv[PGP_AEAD_MAX_NONCE_LEN]; /* iv for AEAD mode */
     uint8_t                 ad[PGP_AEAD_MAX_AD_LEN];    /* additional data for AEAD mode */
@@ -115,13 +114,13 @@ typedef struct pgp_dest_signed_param_t {
     rnp_ctx_t *              ctx;      /* rnp operation context with additional parameters */
     pgp_password_provider_t *password_provider;   /* password provider from write handler */
     std::vector<pgp_dest_signer_info_t> siginfos; /* list of  pgp_dest_signer_info_t */
-    std::vector<pgp_hash_t> hashes;    /* hashes to pass raw data through and then sign */
-    bool                    clr_start; /* we are on the start of the line */
-    uint8_t                 clr_buf[CT_BUF_LEN]; /* buffer to hold partial line data */
-    size_t                  clr_buflen;          /* number of bytes in buffer */
+    rnp::HashList hashes;              /* hashes to pass raw data through and then sign */
+    bool          clr_start;           /* we are on the start of the line */
+    uint8_t       clr_buf[CT_BUF_LEN]; /* buffer to hold partial line data */
+    size_t        clr_buflen;          /* number of bytes in buffer */
 
     pgp_dest_signed_param_t() = default;
-    ~pgp_dest_signed_param_t();
+    ~pgp_dest_signed_param_t() = default;
 } pgp_dest_signed_param_t;
 
 typedef struct pgp_dest_partial_param_t {
@@ -298,7 +297,12 @@ encrypted_dst_write_cfb(pgp_dest_t *dst, const void *buf, size_t len)
     }
 
     if (param->has_mdc) {
-        pgp_hash_add(&param->mdc, buf, len);
+        try {
+            param->mdc.add(buf, len);
+        } catch (const std::exception &e) {
+            RNP_LOG("%s", e.what());
+            return RNP_ERROR_BAD_STATE;
+        }
     }
 
     while (len > 0) {
@@ -312,6 +316,7 @@ encrypted_dst_write_cfb(pgp_dest_t *dst, const void *buf, size_t len)
     return RNP_SUCCESS;
 }
 
+#if defined(ENABLE_AEAD)
 static rnp_result_t
 encrypted_start_aead_chunk(pgp_dest_encrypted_param_t *param, size_t idx, bool last)
 {
@@ -383,10 +388,15 @@ encrypted_start_aead_chunk(pgp_dest_encrypted_param_t *param, size_t idx, bool l
 
     return res ? RNP_SUCCESS : RNP_ERROR_BAD_PARAMETERS;
 }
+#endif
 
 static rnp_result_t
 encrypted_dst_write_aead(pgp_dest_t *dst, const void *buf, size_t len)
 {
+#if !defined(ENABLE_AEAD)
+    RNP_LOG("AEAD is not enabled.");
+    return RNP_ERROR_WRITE;
+#else
     pgp_dest_encrypted_param_t *param = (pgp_dest_encrypted_param_t *) dst->param;
 
     size_t       sz;
@@ -439,33 +449,43 @@ encrypted_dst_write_aead(pgp_dest_t *dst, const void *buf, size_t len)
     }
 
     return RNP_SUCCESS;
+#endif
 }
 
 static rnp_result_t
 encrypted_dst_finish(pgp_dest_t *dst)
 {
-    uint8_t                     mdcbuf[MDC_V1_SIZE];
     pgp_dest_encrypted_param_t *param = (pgp_dest_encrypted_param_t *) dst->param;
-    rnp_result_t                res;
 
     if (param->aead) {
+#if !defined(ENABLE_AEAD)
+        RNP_LOG("AEAD is not enabled.");
+        rnp_result_t res = RNP_ERROR_NOT_IMPLEMENTED;
+#else
         size_t chunks = param->chunkidx;
         /* if we didn't write anything in current chunk then discard it and restart */
         if (param->chunkout || param->cachelen) {
             chunks++;
         }
 
-        res = encrypted_start_aead_chunk(param, chunks, true);
+        rnp_result_t res = encrypted_start_aead_chunk(param, chunks, true);
         pgp_cipher_aead_destroy(&param->encrypt);
-
+#endif
         if (res) {
+            finish_streamed_packet(&param->pkt);
             return res;
         }
     } else if (param->has_mdc) {
+        uint8_t mdcbuf[MDC_V1_SIZE];
         mdcbuf[0] = MDC_PKT_TAG;
         mdcbuf[1] = MDC_V1_SIZE - 2;
-        pgp_hash_add(&param->mdc, mdcbuf, 2);
-        pgp_hash_finish(&param->mdc, &mdcbuf[2]);
+        try {
+            param->mdc.add(mdcbuf, 2);
+            param->mdc.finish(&mdcbuf[2]);
+        } catch (const std::exception &e) {
+            RNP_LOG("%s", e.what());
+            return RNP_ERROR_BAD_STATE;
+        }
         pgp_cipher_cfb_encrypt(&param->encrypt, mdcbuf, mdcbuf, MDC_V1_SIZE);
         dst_write(param->pkt.writedst, mdcbuf, MDC_V1_SIZE);
     }
@@ -482,14 +502,15 @@ encrypted_dst_close(pgp_dest_t *dst, bool discard)
         return;
     }
 
-    if (!param->aead) {
-        pgp_hash_finish(&param->mdc, NULL);
-        pgp_cipher_cfb_finish(&param->encrypt);
-    } else {
+    if (param->aead) {
+#if defined(ENABLE_AEAD)
         pgp_cipher_aead_destroy(&param->encrypt);
+#endif
+    } else {
+        pgp_cipher_cfb_finish(&param->encrypt);
     }
     close_streamed_packet(&param->pkt, discard);
-    free(param);
+    delete param;
     dst->param = NULL;
 }
 
@@ -500,8 +521,6 @@ encrypted_add_recipient(pgp_write_handler_t *handler,
                         const uint8_t *      key,
                         const unsigned       keylen)
 {
-    uint8_t                     enckey[PGP_MAX_KEY_SIZE + 3];
-    unsigned                    checksum = 0;
     pgp_pk_sesskey_t            pkey;
     pgp_dest_encrypted_param_t *param = (pgp_dest_encrypted_param_t *) dst->param;
     rnp_result_t                ret = RNP_ERROR_GENERIC;
@@ -523,89 +542,99 @@ encrypted_add_recipient(pgp_write_handler_t *handler,
     pkey.key_id = userkey->keyid();
 
     /* Encrypt the session key */
+    rnp::secure_array<uint8_t, PGP_MAX_KEY_SIZE + 3> enckey;
     enckey[0] = param->ctx->ealg;
     memcpy(&enckey[1], key, keylen);
 
     /* Calculate checksum */
+    rnp::secure_array<unsigned, 1> checksum;
+
     for (unsigned i = 1; i <= keylen; i++) {
-        checksum += enckey[i];
+        checksum[0] += enckey[i];
     }
-    enckey[keylen + 1] = (checksum >> 8) & 0xff;
-    enckey[keylen + 2] = checksum & 0xff;
+    enckey[keylen + 1] = (checksum[0] >> 8) & 0xff;
+    enckey[keylen + 2] = checksum[0] & 0xff;
 
     pgp_encrypted_material_t material;
 
     switch (userkey->alg()) {
     case PGP_PKA_RSA:
     case PGP_PKA_RSA_ENCRYPT_ONLY: {
-        ret = rsa_encrypt_pkcs1(rnp_ctx_rng_handle(handler->ctx),
+        ret = rsa_encrypt_pkcs1(&handler->ctx->ctx->rng,
                                 &material.rsa,
-                                enckey,
+                                enckey.data(),
                                 keylen + 3,
                                 &userkey->material().rsa);
         if (ret) {
             RNP_LOG("rsa_encrypt_pkcs1 failed");
-            goto finish;
+            return ret;
         }
         break;
     }
     case PGP_PKA_SM2: {
-        ret = sm2_encrypt(rnp_ctx_rng_handle(handler->ctx),
+#if defined(ENABLE_SM2)
+        ret = sm2_encrypt(&handler->ctx->ctx->rng,
                           &material.sm2,
-                          enckey,
+                          enckey.data(),
                           keylen + 3,
                           PGP_HASH_SM3,
                           &userkey->material().ec);
-        if (ret != RNP_SUCCESS) {
+        if (ret) {
             RNP_LOG("sm2_encrypt failed");
-            goto finish;
+            return ret;
         }
         break;
+#else
+        RNP_LOG("sm2_encrypt is not available");
+        return RNP_ERROR_NOT_IMPLEMENTED;
+#endif
     }
     case PGP_PKA_ECDH: {
-        ret = ecdh_encrypt_pkcs5(rnp_ctx_rng_handle(handler->ctx),
+        if (!curve_supported(userkey->material().ec.curve)) {
+            RNP_LOG("ECDH encrypt: curve %d is not supported.",
+                    (int) userkey->material().ec.curve);
+            return RNP_ERROR_NOT_SUPPORTED;
+        }
+        ret = ecdh_encrypt_pkcs5(&handler->ctx->ctx->rng,
                                  &material.ecdh,
-                                 enckey,
+                                 enckey.data(),
                                  keylen + 3,
                                  &userkey->material().ec,
                                  userkey->fp());
-        if (ret != RNP_SUCCESS) {
+        if (ret) {
             RNP_LOG("ECDH encryption failed %d", ret);
-            goto finish;
+            return ret;
         }
         break;
     }
     case PGP_PKA_ELGAMAL: {
-        ret = elgamal_encrypt_pkcs1(rnp_ctx_rng_handle(handler->ctx),
+        ret = elgamal_encrypt_pkcs1(&handler->ctx->ctx->rng,
                                     &material.eg,
-                                    enckey,
+                                    enckey.data(),
                                     keylen + 3,
                                     &userkey->material().eg);
         if (ret) {
             RNP_LOG("pgp_elgamal_public_encrypt failed");
-            goto finish;
+            return ret;
         }
         break;
     }
     default:
         RNP_LOG("unsupported alg: %d", (int) userkey->alg());
-        goto finish;
+        return ret;
     }
 
     /* Writing symmetric key encrypted session key packet */
     try {
         pkey.write_material(material);
         pkey.write(*param->pkt.origdst);
-        ret = param->pkt.origdst->werr;
+        return param->pkt.origdst->werr;
     } catch (const std::exception &e) {
-        ret = RNP_ERROR_WRITE;
+        return RNP_ERROR_WRITE;
     }
-finish:
-    pgp_forget(enckey, sizeof(enckey));
-    pgp_forget(&checksum, sizeof(checksum));
-    return ret;
 }
 
+#if defined(ENABLE_AEAD)
 static bool
 encrypted_sesk_set_ad(pgp_crypt_t *crypt, pgp_sk_sesskey_t *skey)
 {
@@ -618,6 +647,7 @@ encrypted_sesk_set_ad(pgp_crypt_t *crypt, pgp_sk_sesskey_t *skey)
 
     return pgp_cipher_aead_set_ad(crypt, ad_data, 4);
 }
+#endif
 
 static rnp_result_t
 encrypted_add_password(rnp_symmetric_pass_info_t * pass,
@@ -629,8 +659,6 @@ encrypted_add_password(rnp_symmetric_pass_info_t * pass,
     pgp_sk_sesskey_t skey = {};
     unsigned         s2keylen; /* length of the s2k key */
     pgp_crypt_t      kcrypt;
-    uint8_t          nonce[PGP_AEAD_MAX_NONCE_LEN];
-    bool             res;
 
     skey.alg = param->ctx->ealg;
     skey.s2k = pass->s2k;
@@ -647,14 +675,14 @@ encrypted_add_password(rnp_symmetric_pass_info_t * pass,
         if (singlepass) {
             /* if there are no public keys then we do not encrypt session key in the packet */
             skey.enckeylen = 0;
-            memcpy(key, pass->key, s2keylen);
+            memcpy(key, pass->key.data(), s2keylen);
         } else {
             /* Currently we are using the same sym algo for key and stream encryption */
             skey.enckeylen = keylen + 1;
             skey.enckey[0] = param->ctx->ealg;
             memcpy(&skey.enckey[1], key, keylen);
             skey.alg = pass->s2k_cipher;
-            if (!pgp_cipher_cfb_start(&kcrypt, skey.alg, pass->key, NULL)) {
+            if (!pgp_cipher_cfb_start(&kcrypt, skey.alg, pass->key.data(), NULL)) {
                 RNP_LOG("key encryption failed");
                 return RNP_ERROR_BAD_PARAMETERS;
             }
@@ -662,6 +690,10 @@ encrypted_add_password(rnp_symmetric_pass_info_t * pass,
             pgp_cipher_cfb_finish(&kcrypt);
         }
     } else {
+#if !defined(ENABLE_AEAD)
+        RNP_LOG("AEAD support is not enabled.");
+        return RNP_ERROR_NOT_IMPLEMENTED;
+#else
         /* AEAD-encrypted v5 packet */
         if ((param->ctx->aalg != PGP_AEAD_EAX) && (param->ctx->aalg != PGP_AEAD_OCB)) {
             RNP_LOG("unsupported AEAD algorithm");
@@ -673,12 +705,14 @@ encrypted_add_password(rnp_symmetric_pass_info_t * pass,
         skey.ivlen = pgp_cipher_aead_nonce_len(skey.aalg);
         skey.enckeylen = keylen + pgp_cipher_aead_tag_len(skey.aalg);
 
-        if (!rng_get_data(rnp_ctx_rng_handle(param->ctx), skey.iv, skey.ivlen)) {
+        try {
+            param->ctx->ctx->rng.get(skey.iv, skey.ivlen);
+        } catch (const std::exception &e) {
             return RNP_ERROR_RNG;
         }
 
         /* initialize cipher */
-        if (!pgp_cipher_aead_init(&kcrypt, skey.alg, skey.aalg, pass->key, false)) {
+        if (!pgp_cipher_aead_init(&kcrypt, skey.alg, skey.aalg, pass->key.data(), false)) {
             return RNP_ERROR_BAD_PARAMETERS;
         }
 
@@ -688,17 +722,19 @@ encrypted_add_password(rnp_symmetric_pass_info_t * pass,
         }
 
         /* calculate nonce */
-        size_t nlen = pgp_cipher_aead_nonce(skey.aalg, skey.iv, nonce, 0);
+        uint8_t nonce[PGP_AEAD_MAX_NONCE_LEN];
+        size_t  nlen = pgp_cipher_aead_nonce(skey.aalg, skey.iv, nonce, 0);
 
         /* start cipher, encrypt key and get tag */
-        res = pgp_cipher_aead_start(&kcrypt, nonce, nlen) &&
-              pgp_cipher_aead_finish(&kcrypt, skey.enckey, key, keylen);
+        bool res = pgp_cipher_aead_start(&kcrypt, nonce, nlen) &&
+                   pgp_cipher_aead_finish(&kcrypt, skey.enckey, key, keylen);
 
         pgp_cipher_aead_destroy(&kcrypt);
 
         if (!res) {
             return RNP_ERROR_BAD_STATE;
         }
+#endif
     }
 
     /* Writing symmetric key encrypted session key packet */
@@ -721,8 +757,10 @@ encrypted_start_cfb(pgp_dest_encrypted_param_t *param, uint8_t *enckey)
         /* initializing the mdc */
         dst_write(param->pkt.writedst, &mdcver, 1);
 
-        if (!pgp_hash_create(&param->mdc, PGP_HASH_SHA1)) {
-            RNP_LOG("cannot create sha1 hash");
+        try {
+            param->mdc = rnp::Hash(PGP_HASH_SHA1);
+        } catch (const std::exception &e) {
+            RNP_LOG("cannot create sha1 hash: %s", e.what());
             return RNP_ERROR_GENERIC;
         }
     }
@@ -734,15 +772,17 @@ encrypted_start_cfb(pgp_dest_encrypted_param_t *param, uint8_t *enckey)
 
     /* generating and writing iv/password check bytes */
     blsize = pgp_block_size(param->ctx->ealg);
-    if (!rng_get_data(rnp_ctx_rng_handle(param->ctx), enchdr, blsize)) {
-        return RNP_ERROR_RNG;
-    }
+    try {
+        param->ctx->ctx->rng.get(enchdr, blsize);
+        enchdr[blsize] = enchdr[blsize - 2];
+        enchdr[blsize + 1] = enchdr[blsize - 1];
 
-    enchdr[blsize] = enchdr[blsize - 2];
-    enchdr[blsize + 1] = enchdr[blsize - 1];
-
-    if (param->has_mdc) {
-        pgp_hash_add(&param->mdc, enchdr, blsize + 2);
+        if (param->has_mdc) {
+            param->mdc.add(enchdr, blsize + 2);
+        }
+    } catch (const std::exception &e) {
+        RNP_LOG("%s", e.what());
+        return RNP_ERROR_BAD_STATE;
     }
 
     pgp_cipher_cfb_encrypt(&param->encrypt, enchdr, enchdr, blsize + 2);
@@ -761,6 +801,10 @@ encrypted_start_cfb(pgp_dest_encrypted_param_t *param, uint8_t *enckey)
 static rnp_result_t
 encrypted_start_aead(pgp_dest_encrypted_param_t *param, uint8_t *enckey)
 {
+#if !defined(ENABLE_AEAD)
+    RNP_LOG("AEAD support is not enabled.");
+    return RNP_ERROR_NOT_IMPLEMENTED;
+#else
     uint8_t hdr[4 + PGP_AEAD_MAX_NONCE_LEN];
     size_t  nlen;
 
@@ -776,7 +820,9 @@ encrypted_start_aead(pgp_dest_encrypted_param_t *param, uint8_t *enckey)
 
     /* generate iv */
     nlen = pgp_cipher_aead_nonce_len(param->ctx->aalg);
-    if (!rng_get_data(rnp_ctx_rng_handle(param->ctx), param->iv, nlen)) {
+    try {
+        param->ctx->ctx->rng.get(param->iv, nlen);
+    } catch (const std::exception &e) {
         return RNP_ERROR_RNG;
     }
     memcpy(hdr + 4, param->iv, nlen);
@@ -801,6 +847,7 @@ encrypted_start_aead(pgp_dest_encrypted_param_t *param, uint8_t *enckey)
     }
 
     return encrypted_start_aead_chunk(param, 0, false);
+#endif
 }
 
 static rnp_result_t
@@ -810,7 +857,6 @@ init_encrypted_dst(pgp_write_handler_t *handler, pgp_dest_t *dst, pgp_dest_t *wr
     bool                        singlepass = true;
     unsigned                    pkeycount = 0;
     unsigned                    skeycount = 0;
-    uint8_t                     enckey[PGP_MAX_KEY_SIZE] = {0}; /* content encryption key */
     unsigned                    keylen;
     rnp_result_t                ret = RNP_ERROR_GENERIC;
 
@@ -837,11 +883,16 @@ init_encrypted_dst(pgp_write_handler_t *handler, pgp_dest_t *dst, pgp_dest_t *wr
         }
     }
 
-    if (!init_dst_common(dst, sizeof(*param))) {
+    if (!init_dst_common(dst, 0)) {
+        return RNP_ERROR_OUT_OF_MEMORY;
+    }
+    try {
+        param = new pgp_dest_encrypted_param_t();
+        dst->param = param;
+    } catch (const std::exception &e) {
+        RNP_LOG("%s", e.what());
         return RNP_ERROR_OUT_OF_MEMORY;
     }
-
-    param = (pgp_dest_encrypted_param_t *) dst->param;
     param->has_mdc = true;
     param->aead = handler->ctx->aalg != PGP_AEAD_NONE;
     param->aalg = handler->ctx->aalg;
@@ -855,6 +906,7 @@ init_encrypted_dst(pgp_write_handler_t *handler, pgp_dest_t *dst, pgp_dest_t *wr
     pkeycount = handler->ctx->recipients.size();
     skeycount = handler->ctx->passwords.size();
 
+    rnp::secure_array<uint8_t, PGP_MAX_KEY_SIZE> enckey; /* content encryption key */
     if (!pkeycount && !skeycount) {
         RNP_LOG("no recipients");
         ret = RNP_ERROR_BAD_PARAMETERS;
@@ -862,7 +914,9 @@ init_encrypted_dst(pgp_write_handler_t *handler, pgp_dest_t *dst, pgp_dest_t *wr
     }
 
     if ((pkeycount > 0) || (skeycount > 1) || param->aead) {
-        if (!rng_get_data(rnp_ctx_rng_handle(handler->ctx), enckey, keylen)) {
+        try {
+            handler->ctx->ctx->rng.get(enckey.data(), keylen);
+        } catch (const std::exception &e) {
             ret = RNP_ERROR_RNG;
             goto finish;
         }
@@ -871,7 +925,7 @@ init_encrypted_dst(pgp_write_handler_t *handler, pgp_dest_t *dst, pgp_dest_t *wr
 
     /* Configuring and writing pk-encrypted session keys */
     for (auto recipient : handler->ctx->recipients) {
-        ret = encrypted_add_recipient(handler, dst, recipient, enckey, keylen);
+        ret = encrypted_add_recipient(handler, dst, recipient, enckey.data(), keylen);
         if (ret) {
             goto finish;
         }
@@ -879,7 +933,7 @@ init_encrypted_dst(pgp_write_handler_t *handler, pgp_dest_t *dst, pgp_dest_t *wr
 
     /* Configuring and writing sk-encrypted session key(s) */
     for (auto &passinfo : handler->ctx->passwords) {
-        ret = encrypted_add_password(&passinfo, param, enckey, keylen, singlepass);
+        ret = encrypted_add_password(&passinfo, param, enckey.data(), keylen, singlepass);
         if (ret != RNP_SUCCESS) {
             goto finish;
         }
@@ -904,14 +958,13 @@ init_encrypted_dst(pgp_write_handler_t *handler, pgp_dest_t *dst, pgp_dest_t *wr
 
     if (param->aead) {
         /* initialize AEAD encryption */
-        ret = encrypted_start_aead(param, enckey);
+        ret = encrypted_start_aead(param, enckey.data());
     } else {
         /* initialize old CFB or CFB with MDC */
-        ret = encrypted_start_cfb(param, enckey);
+        ret = encrypted_start_cfb(param, enckey.data());
     }
 finish:
     handler->ctx->passwords.clear();
-    pgp_forget(enckey, sizeof(enckey));
     if (ret) {
         encrypted_dst_close(dst, true);
     }
@@ -943,29 +996,33 @@ cleartext_dst_writeline(pgp_dest_signed_param_t *param,
     /* output data */
     dst_write(param->writedst, buf, len);
 
-    if (eol) {
-        bool hashcrlf = false;
-        ptr = buf + len - 1;
+    try {
+        if (eol) {
+            bool hashcrlf = false;
+            ptr = buf + len - 1;
 
-        /* skipping trailing characters - space, tab, carriage return, line feed */
-        while ((ptr >= buf) && ((*ptr == CH_SPACE) || (*ptr == CH_TAB) || (*ptr == CH_CR) ||
-                                (*ptr == CH_LF))) {
-            if (*ptr == CH_LF) {
-                hashcrlf = true;
+            /* skipping trailing characters - space, tab, carriage return, line feed */
+            while ((ptr >= buf) && ((*ptr == CH_SPACE) || (*ptr == CH_TAB) ||
+                                    (*ptr == CH_CR) || (*ptr == CH_LF))) {
+                if (*ptr == CH_LF) {
+                    hashcrlf = true;
+                }
+                ptr--;
             }
-            ptr--;
-        }
 
-        /* hashing line body and \r\n */
-        pgp_hash_list_update(param->hashes, buf, ptr + 1 - buf);
-        if (hashcrlf) {
-            pgp_hash_list_update(param->hashes, ST_CRLF, 2);
+            /* hashing line body and \r\n */
+            param->hashes.add(buf, ptr + 1 - buf);
+            if (hashcrlf) {
+                param->hashes.add(ST_CRLF, 2);
+            }
+            param->clr_start = hashcrlf;
+        } else if (len > 0) {
+            /* hashing just line's data */
+            param->hashes.add(buf, len);
+            param->clr_start = false;
         }
-        param->clr_start = hashcrlf;
-    } else if (len > 0) {
-        /* hashing just line's data */
-        pgp_hash_list_update(param->hashes, buf, len);
-        param->clr_start = false;
+    } catch (const std::exception &e) {
+        RNP_LOG("failed to hash data: %s", e.what());
     }
 }
 
@@ -1049,56 +1106,35 @@ cleartext_dst_write(pgp_dest_t *dst, const void *buf, size_t len)
     return RNP_SUCCESS;
 }
 
-static rnp_result_t
-signed_fill_signature(pgp_dest_signed_param_t *param,
-                      pgp_signature_t *        sig,
-                      pgp_dest_signer_info_t * signer)
+static void
+signed_fill_signature(pgp_dest_signed_param_t &param,
+                      pgp_signature_t &        sig,
+                      pgp_dest_signer_info_t & signer)
 {
-    const pgp_key_pkt_t *deckey = NULL;
-    pgp_hash_t           hash;
-    pgp_password_ctx_t   ctx = {.op = PGP_OP_SIGN, .key = signer->key};
-    rnp_result_t         ret = RNP_ERROR_GENERIC;
-
-    /* fill signature fields */
-    try {
-        sig->set_keyfp(signer->key->fp());
-        sig->set_keyid(signer->key->keyid());
-        sig->set_creation(signer->sigcreate ? signer->sigcreate : time(NULL));
-        sig->set_expiration(signer->sigexpire);
-    } catch (const std::exception &e) {
-        RNP_LOG("failed to setup signature fields: %s", e.what());
-        return RNP_ERROR_OUT_OF_MEMORY;
-    }
-    if (!signature_fill_hashed_data(sig)) {
-        RNP_LOG("failed to fill the signature data");
-        return RNP_ERROR_OUT_OF_MEMORY;
+    /* fill signature fields, assuming sign_init was called on it */
+    if (signer.sigcreate) {
+        sig.set_creation(signer.sigcreate);
     }
+    sig.set_expiration(signer.sigexpire);
+    sig.fill_hashed_data();
 
-    if (!pgp_hash_copy(&hash, pgp_hash_list_get(param->hashes, sig->halg))) {
+    const rnp::Hash *listh = param.hashes.get(sig.halg);
+    if (!listh) {
         RNP_LOG("failed to obtain hash");
-        return RNP_ERROR_BAD_PARAMETERS;
+        throw rnp::rnp_exception(RNP_ERROR_BAD_STATE);
     }
+    rnp::Hash hash;
+    listh->clone(hash);
 
     /* decrypt the secret key if needed */
-    if (signer->key->encrypted()) {
-        deckey = pgp_decrypt_seckey(signer->key, param->password_provider, &ctx);
-        if (!deckey) {
-            RNP_LOG("wrong secret key password");
-            pgp_hash_finish(&hash, NULL);
-            return RNP_ERROR_BAD_PASSWORD;
-        }
-    } else {
-        deckey = &signer->key->pkt();
+    rnp::KeyLocker(*signer.key);
+    if (signer.key->encrypted() &&
+        !signer.key->unlock(*param.password_provider, PGP_OP_SIGN)) {
+        RNP_LOG("wrong secret key password");
+        throw rnp::rnp_exception(RNP_ERROR_BAD_PASSWORD);
     }
-
     /* calculate the signature */
-    ret = signature_calculate(sig, &deckey->material, &hash, rnp_ctx_rng_handle(param->ctx));
-
-    /* destroy decrypted secret key */
-    if (signer->key->encrypted()) {
-        delete deckey;
-    }
-    return ret;
+    signature_calculate(sig, signer.key->material(), hash, *param.ctx->ctx);
 }
 
 static rnp_result_t
@@ -1106,27 +1142,24 @@ signed_write_signature(pgp_dest_signed_param_t *param,
                        pgp_dest_signer_info_t * signer,
                        pgp_dest_t *             writedst)
 {
-    pgp_signature_t sig;
-    sig.version = (pgp_version_t) 4;
-    if (signer->onepass.version) {
-        sig.halg = signer->onepass.halg;
-        sig.palg = signer->onepass.palg;
-        sig.set_type(signer->onepass.type);
-    } else {
-        sig.halg = pgp_hash_adjust_alg_to_key(signer->halg, &signer->key->pkt());
-        sig.palg = signer->key->alg();
-        sig.set_type(param->ctx->detached ? PGP_SIG_BINARY : PGP_SIG_TEXT);
-    }
-
-    rnp_result_t ret = signed_fill_signature(param, &sig, signer);
-    if (ret) {
-        return ret;
-    }
     try {
+        pgp_signature_t sig;
+        if (signer->onepass.version) {
+            signer->key->sign_init(sig, signer->onepass.halg);
+            sig.palg = signer->onepass.palg;
+            sig.set_type(signer->onepass.type);
+        } else {
+            signer->key->sign_init(sig, signer->halg);
+            /* line below should be checked */
+            sig.set_type(param->ctx->detached ? PGP_SIG_BINARY : PGP_SIG_TEXT);
+        }
+        signed_fill_signature(*param, sig, *signer);
         sig.write(*writedst);
         return writedst->werr;
+    } catch (const rnp::rnp_exception &e) {
+        return e.code();
     } catch (const std::exception &e) {
-        RNP_LOG("%s", e.what());
+        RNP_LOG("Failed to write signature: %s", e.what());
         return RNP_ERROR_WRITE;
     }
 }
@@ -1213,7 +1246,7 @@ static void
 signed_dst_update(pgp_dest_t *dst, const void *buf, size_t len)
 {
     pgp_dest_signed_param_t *param = (pgp_dest_signed_param_t *) dst->param;
-    pgp_hash_list_update(param->hashes, buf, len);
+    param->hashes.add(buf, len);
 }
 
 static rnp_result_t
@@ -1225,6 +1258,12 @@ signed_add_signer(pgp_dest_signed_param_t *param, rnp_signer_info_t *signer, boo
         RNP_LOG("secret key required for signing");
         return RNP_ERROR_BAD_PARAMETERS;
     }
+    /* validate signing key material if didn't before */
+    signer->key->pkt().material.validate(*param->ctx->ctx, false);
+    if (!signer->key->pkt().material.valid()) {
+        RNP_LOG("attempt to sign to the key with invalid material");
+        return RNP_ERROR_NO_SUITABLE_KEY;
+    }
 
     /* copy fields */
     sinfo.key = signer->key;
@@ -1233,7 +1272,10 @@ signed_add_signer(pgp_dest_signed_param_t *param, rnp_signer_info_t *signer, boo
 
     /* Add hash to the list */
     sinfo.halg = pgp_hash_adjust_alg_to_key(signer->halg, &signer->key->pkt());
-    if (!pgp_hash_list_add(param->hashes, sinfo.halg)) {
+    try {
+        param->hashes.add_alg(sinfo.halg);
+    } catch (const std::exception &e) {
+        RNP_LOG("%s", e.what());
         return RNP_ERROR_BAD_PARAMETERS;
     }
 
@@ -1279,19 +1321,11 @@ signed_add_signer(pgp_dest_signed_param_t *param, rnp_signer_info_t *signer, boo
     }
 }
 
-pgp_dest_signed_param_t::~pgp_dest_signed_param_t()
-{
-    for (auto &hash : hashes) {
-        pgp_hash_finish(&hash, NULL);
-    }
-}
-
 static rnp_result_t
 init_signed_dst(pgp_write_handler_t *handler, pgp_dest_t *dst, pgp_dest_t *writedst)
 {
     pgp_dest_signed_param_t *param;
     rnp_result_t             ret = RNP_ERROR_GENERIC;
-    const char *             hname;
 
     if (!handler->key_provider) {
         RNP_LOG("no key provider");
@@ -1345,10 +1379,10 @@ init_signed_dst(pgp_write_handler_t *handler, pgp_dest_t *dst, pgp_dest_t *write
         dst_write(param->writedst, ST_CRLF, strlen(ST_CRLF));
         dst_write(param->writedst, ST_HEADER_HASH, strlen(ST_HEADER_HASH));
 
-        for (const auto &hash : param->hashes) {
-            hname = pgp_hash_name(&hash);
+        for (const auto &hash : param->hashes.hashes()) {
+            auto hname = rnp::Hash::name(hash.alg());
             dst_write(param->writedst, hname, strlen(hname));
-            if (&hash != &param->hashes.back()) {
+            if (&hash != &param->hashes.hashes().back()) {
                 dst_write(param->writedst, ST_COMMA, 1);
             }
         }
@@ -1969,16 +2003,15 @@ done:
 }
 
 rnp_result_t
-rnp_raw_encrypt_src(pgp_source_t &src, pgp_dest_t &dst, const std::string &password)
+rnp_raw_encrypt_src(pgp_source_t &        src,
+                    pgp_dest_t &          dst,
+                    const std::string &   password,
+                    rnp::SecurityContext &secctx)
 {
     pgp_write_handler_t handler = {};
     rnp_ctx_t           ctx;
-    rng_t               rng = {};
 
-    if (!rng_init(&rng, RNG_SYSTEM)) {
-        return RNP_ERROR_BAD_STATE;
-    }
-    ctx.rng = &rng;
+    ctx.ctx = &secctx;
     ctx.ealg = DEFAULT_PGP_SYMM_ALG;
     handler.ctx = &ctx;
     pgp_dest_t encrypted = {};
@@ -1997,6 +2030,5 @@ rnp_raw_encrypt_src(pgp_source_t &src, pgp_dest_t &dst, const std::string &passw
     ret = dst_write_src(&src, &encrypted);
 done:
     dst_close(&encrypted, ret);
-    rng_destroy(&rng);
     return ret;
 }
