@@ -166,12 +166,38 @@ TEST_F(rnp_tests, test_ffi_encrypt_pass)
     assert_rnp_success(rnp_output_to_path(&output, "encrypted"));
     assert_non_null(output);
     // create encrypt operation
+    assert_rnp_failure(rnp_op_encrypt_create(NULL, ffi, input, output));
+    assert_rnp_failure(rnp_op_encrypt_create(&op, NULL, input, output));
+    assert_rnp_failure(rnp_op_encrypt_create(&op, ffi, NULL, output));
+    assert_rnp_failure(rnp_op_encrypt_create(&op, ffi, input, NULL));
     assert_rnp_success(rnp_op_encrypt_create(&op, ffi, input, output));
     // add password (using all defaults)
+    assert_rnp_failure(rnp_op_encrypt_add_password(NULL, "pass1", NULL, 0, NULL));
+    assert_rnp_failure(rnp_op_encrypt_add_password(op, "", NULL, 0, NULL));
+    assert_rnp_failure(rnp_op_encrypt_add_password(op, "pass1", "WRONG", 0, NULL));
+    assert_rnp_failure(rnp_op_encrypt_add_password(op, "pass1", NULL, 0, "WRONG"));
     assert_rnp_success(rnp_op_encrypt_add_password(op, "pass1", NULL, 0, NULL));
     // add password
-    assert_rnp_success(rnp_op_encrypt_add_password(op, "pass2", "SM3", 12345, "TWOFISH"));
+    if (!sm2_enabled() && !twofish_enabled()) {
+        assert_rnp_failure(rnp_op_encrypt_add_password(op, "pass2", "SM3", 12345, "TWOFISH"));
+        assert_rnp_failure(
+          rnp_op_encrypt_add_password(op, "pass2", "SHA256", 12345, "TWOFISH"));
+        assert_rnp_success(
+          rnp_op_encrypt_add_password(op, "pass2", "SHA256", 12345, "BLOWFISH"));
+    } else if (!sm2_enabled() && twofish_enabled()) {
+        assert_rnp_failure(rnp_op_encrypt_add_password(op, "pass2", "SM3", 12345, "TWOFISH"));
+        assert_rnp_success(
+          rnp_op_encrypt_add_password(op, "pass2", "SHA256", 12345, "TWOFISH"));
+    } else if (sm2_enabled() && !twofish_enabled()) {
+        assert_rnp_failure(rnp_op_encrypt_add_password(op, "pass2", "SM3", 12345, "TWOFISH"));
+        assert_rnp_success(rnp_op_encrypt_add_password(op, "pass2", "SM3", 12345, "BLOWFISH"));
+    } else {
+        assert_rnp_success(rnp_op_encrypt_add_password(op, "pass2", "SM3", 12345, "TWOFISH"));
+    }
     // set the data encryption cipher
+    assert_rnp_failure(rnp_op_encrypt_set_cipher(NULL, "CAST5"));
+    assert_rnp_failure(rnp_op_encrypt_set_cipher(op, NULL));
+    assert_rnp_failure(rnp_op_encrypt_set_cipher(op, "WRONG"));
     assert_rnp_success(rnp_op_encrypt_set_cipher(op, "CAST5"));
     // execute the operation
     assert_rnp_success(rnp_op_encrypt_execute(op));
@@ -221,8 +247,7 @@ TEST_F(rnp_tests, test_ffi_encrypt_pass)
     assert_non_null(input);
     assert_rnp_success(rnp_output_to_path(&output, "decrypted"));
     assert_non_null(output);
-    assert_int_equal(
-      RNP_SUCCESS,
+    assert_rnp_success(
       rnp_ffi_set_pass_provider(ffi, ffi_string_password_provider, (void *) "pass1"));
     assert_rnp_success(rnp_decrypt(ffi, input, output));
     // cleanup
@@ -281,7 +306,19 @@ TEST_F(rnp_tests, test_ffi_encrypt_pass_provider)
     assert_rnp_success(rnp_ffi_set_pass_provider(ffi, getpasscb_inc, &pswdnum));
     assert_rnp_success(rnp_op_encrypt_add_password(op, NULL, NULL, 0, NULL));
     // add another password with different encryption parameters
-    assert_rnp_success(rnp_op_encrypt_add_password(op, NULL, "SM3", 12345, "TWOFISH"));
+    if (!sm2_enabled() && !twofish_enabled()) {
+        assert_rnp_failure(rnp_op_encrypt_add_password(op, NULL, "SM3", 12345, "TWOFISH"));
+        assert_rnp_failure(rnp_op_encrypt_add_password(op, NULL, "SHA256", 12345, "TWOFISH"));
+        assert_rnp_success(rnp_op_encrypt_add_password(op, NULL, "SHA256", 12345, "BLOWFISH"));
+    } else if (!sm2_enabled() && twofish_enabled()) {
+        assert_rnp_failure(rnp_op_encrypt_add_password(op, NULL, "SM3", 12345, "TWOFISH"));
+        assert_rnp_success(rnp_op_encrypt_add_password(op, NULL, "SHA256", 12345, "TWOFISH"));
+    } else if (sm2_enabled() && !twofish_enabled()) {
+        assert_rnp_failure(rnp_op_encrypt_add_password(op, NULL, "SM3", 12345, "TWOFISH"));
+        assert_rnp_success(rnp_op_encrypt_add_password(op, NULL, "SM3", 12345, "BLOWFISH"));
+    } else {
+        assert_rnp_success(rnp_op_encrypt_add_password(op, NULL, "SM3", 12345, "TWOFISH"));
+    }
     // set the data encryption cipher
     assert_rnp_success(rnp_op_encrypt_set_cipher(op, "CAMELLIA256"));
     // execute the operation
@@ -355,6 +392,8 @@ TEST_F(rnp_tests, test_ffi_encrypt_pk)
     // add recipients
     rnp_key_handle_t key = NULL;
     assert_rnp_success(rnp_locate_key(ffi, "userid", "key0-uid2", &key));
+    assert_rnp_failure(rnp_op_encrypt_add_recipient(NULL, key));
+    assert_rnp_failure(rnp_op_encrypt_add_recipient(op, NULL));
     assert_rnp_success(rnp_op_encrypt_add_recipient(op, key));
     rnp_key_handle_destroy(key);
     key = NULL;
@@ -594,12 +633,20 @@ TEST_F(rnp_tests, test_ffi_encrypt_and_sign)
     // set the data encryption cipher
     assert_rnp_success(rnp_op_encrypt_set_cipher(op, "CAST5"));
     // enable armoring
+    assert_rnp_failure(rnp_op_encrypt_set_armor(NULL, true));
     assert_rnp_success(rnp_op_encrypt_set_armor(op, true));
     // add signature
+    assert_rnp_failure(rnp_op_encrypt_set_hash(NULL, "SHA256"));
+    assert_rnp_failure(rnp_op_encrypt_set_hash(op, NULL));
+    assert_rnp_failure(rnp_op_encrypt_set_hash(op, "WRONG"));
     assert_rnp_success(rnp_op_encrypt_set_hash(op, "SHA1"));
+    assert_rnp_failure(rnp_op_encrypt_set_creation_time(NULL, 0));
     assert_rnp_success(rnp_op_encrypt_set_creation_time(op, 0));
+    assert_rnp_failure(rnp_op_encrypt_set_expiration_time(NULL, 0));
     assert_rnp_success(rnp_op_encrypt_set_expiration_time(op, 0));
     assert_rnp_success(rnp_locate_key(ffi, "userid", "key1-uid1", &key));
+    assert_rnp_failure(rnp_op_encrypt_add_signature(NULL, key, NULL));
+    assert_rnp_failure(rnp_op_encrypt_add_signature(op, NULL, NULL));
     assert_rnp_success(rnp_op_encrypt_add_signature(op, key, NULL));
     rnp_key_handle_destroy(key);
     key = NULL;
@@ -856,3 +903,67 @@ TEST_F(rnp_tests, test_ffi_encrypt_pk_subkey_selection)
 
     rnp_ffi_destroy(ffi);
 }
+
+TEST_F(rnp_tests, test_ffi_decrypt_small_rsa)
+{
+    rnp_ffi_t   ffi = NULL;
+    const char *plaintext = "data1";
+
+    assert_rnp_success(rnp_ffi_create(&ffi, "GPG", "GPG"));
+    assert_true(import_all_keys(ffi, "data/test_key_validity/rsa_key_small_sig-sec.asc"));
+    rnp_input_t input = NULL;
+    assert_rnp_success(rnp_input_from_path(&input, "data/test_messages/data.enc.small-rsa"));
+    rnp_output_t output = NULL;
+    assert_rnp_success(rnp_output_to_memory(&output, 0));
+    assert_rnp_success(rnp_decrypt(ffi, input, output));
+    size_t   len = 0;
+    uint8_t *buf = NULL;
+    assert_rnp_success(rnp_output_memory_get_buf(output, &buf, &len, false));
+    assert_int_equal(len, 5);
+    assert_int_equal(memcmp(plaintext, buf, 5), 0);
+    assert_rnp_success(rnp_input_destroy(input));
+    assert_rnp_success(rnp_output_destroy(output));
+    rnp_ffi_destroy(ffi);
+}
+
+TEST_F(rnp_tests, test_ffi_decrypt_small_eg)
+{
+    /* make sure unlock and decrypt fails with invalid key */
+    rnp_ffi_t ffi = NULL;
+    assert_rnp_success(rnp_ffi_create(&ffi, "GPG", "GPG"));
+    assert_true(
+      import_all_keys(ffi, "data/test_key_edge_cases/key-eg-small-subgroup-sec.pgp"));
+    rnp_key_handle_t key = NULL;
+    assert_rnp_success(rnp_locate_key(ffi, "keyid", "3b8dda452b9f69b4", &key));
+    assert_non_null(key);
+    /* key is not encrypted */
+    assert_rnp_success(rnp_key_unlock(key, NULL));
+    rnp_key_handle_destroy(key);
+    rnp_input_t input = NULL;
+    assert_rnp_success(
+      rnp_input_from_path(&input, "data/test_messages/message.txt.enc-eg-bad"));
+    rnp_output_t output = NULL;
+    assert_rnp_success(rnp_output_to_null(&output));
+    assert_rnp_failure(rnp_decrypt(ffi, input, output));
+    assert_rnp_success(rnp_input_destroy(input));
+    assert_rnp_success(rnp_output_destroy(output));
+    rnp_ffi_destroy(ffi);
+    /* make sure unlock and decrypt fails with invalid encrypted key */
+    assert_rnp_success(rnp_ffi_create(&ffi, "GPG", "GPG"));
+    assert_rnp_success(
+      rnp_ffi_set_pass_provider(ffi, ffi_string_password_provider, (void *) "password"));
+    assert_true(
+      import_all_keys(ffi, "data/test_key_edge_cases/key-eg-small-subgroup-sec-enc.pgp"));
+    assert_rnp_success(rnp_locate_key(ffi, "keyid", "3b072c3bb2d1a8b2", &key));
+    assert_non_null(key);
+    assert_rnp_success(rnp_key_unlock(key, "password"));
+    assert_rnp_success(rnp_key_lock(key));
+    rnp_key_handle_destroy(key);
+    assert_rnp_success(
+      rnp_input_from_path(&input, "data/test_messages/message.txt.enc-eg-bad2"));
+    assert_rnp_success(rnp_output_to_null(&output));
+    assert_rnp_failure(rnp_decrypt(ffi, input, output));
+    assert_rnp_success(rnp_input_destroy(input));
+    assert_rnp_success(rnp_output_destroy(output));
+    rnp_ffi_destroy(ffi);
+}
